@@ -10,12 +10,11 @@
 
 #[macro_use]
 extern crate bitflags;
+extern crate core_foundation_sys;
+pub extern crate coreaudio_sys as sys;
 
 pub use error::Error;
 
-#[cfg(feature = "audio_toolbox")]
+#[cfg(feature = "audio_unit")]
 pub mod audio_unit;
 pub mod error;
-
-// MacTypes.h
-pub type OSStatus = i32;
