@@ -7,7 +7,7 @@ versa.
 Based on harfbuzz/src/gen-tag-table.py
 
 Input files:
-- https://docs.microsoft.com/en-us/typography/opentype/spec/languagetags (the whole page as HTML)
+- https://learn.microsoft.com/en-us/typography/opentype/spec/languagetags (the whole page as HTML)
 - https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry
 """
 
@@ -690,8 +690,6 @@ ot.add_language('ber', 'BBR')
 ot.remove_language_ot('PGR')
 ot.add_language('el-polyton', 'PGR')
 
-bcp_47.macrolanguages['et'] = {'ekk'}
-
 bcp_47.names['flm'] = 'Falam Chin'
 bcp_47.scopes['flm'] = '(retired code)'
 bcp_47.macrolanguages['flm'] = {'cfm'}
@@ -702,7 +700,6 @@ ot.add_language('und-fonipa', 'IPPH')
 
 ot.add_language('und-fonnapa', 'APPH')
 
-ot.remove_language_ot('IRT')
 ot.add_language('ga-Latg', 'IRT')
 
 ot.add_language('hy-arevmda', 'HYE')
@@ -710,10 +707,6 @@ ot.add_language('hy-arevmda', 'HYE')
 ot.remove_language_ot('KGE')
 ot.add_language('und-Geok', 'KGE')
 
-bcp_47.macrolanguages['id'] = {'in'}
-
-bcp_47.macrolanguages['ijo'] = {'ijc'}
-
 ot.add_language('kht', 'KHN')
 ot.names['KHN'] = ot.names['KHT'] + '(Microsoft fonts)'
 ot.ranks['KHN'] = ot.ranks['KHT'] + 1
@@ -805,8 +798,6 @@ ot.add_language('lzh-Hans', 'ZHS')
 ot.add_language('yue', 'ZHH')
 ot.add_language('yue-Hans', 'ZHS')
 
-bcp_47.macrolanguages['zom'] = {'yos'}
-
 
 def rank_delta(bcp_47, ot):
     """Return a delta to apply to a BCP 47 tag's rank.
