@@ -390,7 +390,7 @@ fn other_error<E>(msg: E) -> std::io::Error
 where
 	E: Into<Box<dyn std::error::Error + Send + Sync>>,
 {
-	std::io::Error::new(std::io::ErrorKind::Other, msg)
+	std::io::Error::other(msg)
 }
 
 impl Settings {
