@@ -4,7 +4,7 @@ mod hint;
 
 use super::{GlyphHMetrics, OutlinePen};
 use hint::{HintParams, HintState, HintingSink};
-use raw::FontRef;
+use raw::{tables::postscript::dict::normalize_font_matrix, FontRef};
 use read_fonts::{
     tables::{
         postscript::{
@@ -143,25 +143,83 @@ impl<'a> Outlines<'a> {
         size: Option<f32>,
         coords: &[F2Dot14],
     ) -> Result<Subfont, Error> {
-        let private_dict_range = self.private_dict_range(index)?;
+        let font_dict = self.parse_font_dict(index)?;
         let blend_state = self
             .top_dict
             .var_store
             .clone()
             .map(|store| BlendState::new(store, coords, 0))
             .transpose()?;
-        let private_dict = PrivateDict::new(self.offset_data, private_dict_range, blend_state)?;
-        let scale = match size {
-            Some(ppem) if self.units_per_em > 0 => {
+        let private_dict =
+            PrivateDict::new(self.offset_data, font_dict.private_dict_range, blend_state)?;
+        let upem = self.units_per_em as i32;
+        let mut scale = match size {
+            Some(ppem) if upem > 0 => {
                 // Note: we do an intermediate scale to 26.6 to ensure we
                 // match FreeType
-                Some(
-                    Fixed::from_bits((ppem * 64.) as i32)
-                        / Fixed::from_bits(self.units_per_em as i32),
-                )
+                Some(Fixed::from_bits((ppem * 64.) as i32) / Fixed::from_bits(upem))
             }
             _ => None,
         };
+        let scale_requested = size.is_some();
+        // Compute our font matrix and adjusted UPEM
+        // See <https://gitlab.freedesktop.org/freetype/freetype/-/blob/f1cd6dbfa0c98f352b698448f40ac27e8fb3832e/src/cff/cffobjs.c#L746>
+        let font_matrix = if let Some((top_matrix, top_upem)) = self.top_dict.font_matrix {
+            // We have a top dict matrix. Now check for a font dict matrix.
+            if let Some((sub_matrix, sub_upem)) = font_dict.font_matrix {
+                let scaling = if top_upem > 1 && sub_upem > 1 {
+                    top_upem.min(sub_upem)
+                } else {
+                    1
+                };
+                // Concatenate and scale
+                let matrix = matrix_mul_scaled(&top_matrix, &sub_matrix, scaling);
+                let upem = Fixed::from_bits(sub_upem)
+                    .mul_div(Fixed::from_bits(top_upem), Fixed::from_bits(scaling));
+                // Then normalize
+                Some(normalize_font_matrix(matrix, upem.to_bits()))
+            } else {
+                // Top matrix was already normalized on load
+                Some((top_matrix, top_upem))
+            }
+        } else if let Some((matrix, upem)) = font_dict.font_matrix {
+            // Just normalize
+            Some(normalize_font_matrix(matrix, upem))
+        } else {
+            None
+        };
+        // Now adjust our scale factor if necessary
+        // See <https://gitlab.freedesktop.org/freetype/freetype/-/blob/f1cd6dbfa0c98f352b698448f40ac27e8fb3832e/src/cff/cffgload.c#L450>
+        let mut font_matrix = if let Some((matrix, matrix_upem)) = font_matrix {
+            // If the scaling factor from our matrix does not equal the nominal
+            // UPEM of the font then adjust the scale.
+            if matrix_upem != upem {
+                // In this case, we need to force a scale for "unscaled"
+                // requests in order to apply the adjusted UPEM from the
+                // font matrix.
+                let original_scale = scale.unwrap_or(Fixed::from_i32(64));
+                scale = Some(
+                    original_scale.mul_div(Fixed::from_bits(upem), Fixed::from_bits(matrix_upem)),
+                );
+            }
+            Some(matrix)
+        } else {
+            None
+        };
+        if font_matrix
+            == Some([
+                Fixed::ONE,
+                Fixed::ZERO,
+                Fixed::ZERO,
+                Fixed::ONE,
+                Fixed::ZERO,
+                Fixed::ZERO,
+            ])
+        {
+            // Let's not waste time applying an identity matrix. This occurs
+            // fairly often after normalization.
+            font_matrix = None;
+        }
         // When hinting, use a modified scale factor
         // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/psft.c#L279>
         let hint_scale = Fixed::from_bits((scale.unwrap_or(Fixed::ONE).to_bits() + 32) / 64);
@@ -169,9 +227,11 @@ impl<'a> Outlines<'a> {
         Ok(Subfont {
             is_cff2: self.is_cff2(),
             scale,
+            scale_requested,
             subrs_offset: private_dict.subrs_offset,
             hint_state,
             store_index: private_dict.store_index,
+            font_matrix,
         })
     }
 
@@ -194,60 +254,90 @@ impl<'a> Outlines<'a> {
         hint: bool,
         pen: &mut impl OutlinePen,
     ) -> Result<(), Error> {
-        let charstring_data = self.top_dict.charstrings.get(glyph_id.to_u32() as usize)?;
+        let cff_data = self.offset_data.as_bytes();
+        let charstrings = self.top_dict.charstrings.clone();
+        let charstring_data = charstrings.get(glyph_id.to_u32() as usize)?;
         let subrs = subfont.subrs(self)?;
         let blend_state = subfont.blend_state(self, coords)?;
+        let cs_eval = CharstringEvaluator {
+            cff_data,
+            charstrings,
+            global_subrs: self.global_subrs.clone(),
+            subrs,
+            blend_state,
+            charstring_data,
+        };
+        // Only apply hinting if we have a scale
+        let apply_hinting = hint && subfont.scale_requested;
         let mut pen_sink = PenSink::new(pen);
         let mut simplifying_adapter = NopFilteringSink::new(&mut pen_sink);
-        // Only apply hinting if we have a scale
-        if hint && subfont.scale.is_some() {
+        if let Some(matrix) = subfont.font_matrix {
+            if apply_hinting {
+                let mut transform_sink =
+                    HintedTransformingSink::new(&mut simplifying_adapter, matrix);
+                let mut hinting_adapter =
+                    HintingSink::new(&subfont.hint_state, &mut transform_sink);
+                cs_eval.evaluate(&mut hinting_adapter)?;
+                hinting_adapter.finish();
+            } else {
+                let mut transform_sink =
+                    ScalingTransformingSink::new(&mut simplifying_adapter, matrix, subfont.scale);
+                cs_eval.evaluate(&mut transform_sink)?;
+            }
+        } else if apply_hinting {
             let mut hinting_adapter =
                 HintingSink::new(&subfont.hint_state, &mut simplifying_adapter);
-            charstring::evaluate(
-                charstring_data,
-                self.global_subrs.clone(),
-                subrs,
-                blend_state,
-                &mut hinting_adapter,
-            )?;
+            cs_eval.evaluate(&mut hinting_adapter)?;
             hinting_adapter.finish();
         } else {
             let mut scaling_adapter =
                 ScalingSink26Dot6::new(&mut simplifying_adapter, subfont.scale);
-            charstring::evaluate(
-                charstring_data,
-                self.global_subrs.clone(),
-                subrs,
-                blend_state,
-                &mut scaling_adapter,
-            )?;
+            cs_eval.evaluate(&mut scaling_adapter)?;
         }
         simplifying_adapter.finish();
         Ok(())
     }
 
-    fn private_dict_range(&self, subfont_index: u32) -> Result<Range<usize>, Error> {
+    fn parse_font_dict(&self, subfont_index: u32) -> Result<FontDict, Error> {
         if self.top_dict.font_dicts.count() != 0 {
             // If we have a font dict array, extract the private dict range
             // from the font dict at the given index.
             let font_dict_data = self.top_dict.font_dicts.get(subfont_index as usize)?;
-            let mut range = None;
-            for entry in dict::entries(font_dict_data, None) {
-                if let dict::Entry::PrivateDictRange(r) = entry? {
-                    range = Some(r);
-                    break;
-                }
-            }
-            range
+            FontDict::new(font_dict_data)
         } else {
             // Use the private dict range from the top dict.
             // Note: "A Private DICT is required but may be specified as having
             // a length of 0 if there are no non-default values to be stored."
             // <https://adobe-type-tools.github.io/font-tech-notes/pdfs/5176.CFF.pdf#page=25>
             let range = self.top_dict.private_dict_range.clone();
-            Some(range.start as usize..range.end as usize)
+            Ok(FontDict {
+                private_dict_range: range.start as usize..range.end as usize,
+                font_matrix: None,
+            })
         }
-        .ok_or(Error::MissingPrivateDict)
+    }
+}
+
+struct CharstringEvaluator<'a> {
+    cff_data: &'a [u8],
+    charstrings: Index<'a>,
+    global_subrs: Index<'a>,
+    subrs: Option<Index<'a>>,
+    blend_state: Option<BlendState<'a>>,
+    charstring_data: &'a [u8],
+}
+
+impl CharstringEvaluator<'_> {
+    fn evaluate(self, sink: &mut impl CommandSink) -> Result<(), Error> {
+        charstring::evaluate(
+            self.cff_data,
+            self.charstrings,
+            self.global_subrs,
+            self.subrs,
+            self.blend_state,
+            self.charstring_data,
+            sink,
+        )
     }
 }
 
@@ -262,9 +352,14 @@ impl<'a> Outlines<'a> {
 pub(crate) struct Subfont {
     is_cff2: bool,
     scale: Option<Fixed>,
+    /// When we have a font matrix, we might force a scale even if the user
+    /// requested unscaled output. In this case, we shouldn't apply hinting
+    /// and this keeps track of that.
+    scale_requested: bool,
     subrs_offset: Option<usize>,
     pub(crate) hint_state: HintState,
     store_index: u16,
+    font_matrix: Option<[Fixed; 6]>,
 }
 
 impl Subfont {
@@ -317,7 +412,7 @@ impl PrivateDict {
                 BlueValues(values) => dict.hint_params.blues = values,
                 FamilyBlues(values) => dict.hint_params.family_blues = values,
                 OtherBlues(values) => dict.hint_params.other_blues = values,
-                FamilyOtherBlues(values) => dict.hint_params.family_blues = values,
+                FamilyOtherBlues(values) => dict.hint_params.family_other_blues = values,
                 BlueScale(value) => dict.hint_params.blue_scale = value,
                 BlueShift(value) => dict.hint_params.blue_shift = value,
                 BlueFuzz(value) => dict.hint_params.blue_fuzz = value,
@@ -339,6 +434,36 @@ impl PrivateDict {
     }
 }
 
+/// Entries that we parse from a Font DICT.
+#[derive(Clone, Default)]
+struct FontDict {
+    private_dict_range: Range<usize>,
+    font_matrix: Option<([Fixed; 6], i32)>,
+}
+
+impl FontDict {
+    fn new(font_dict_data: &[u8]) -> Result<Self, Error> {
+        let mut range = None;
+        let mut font_matrix = None;
+        for entry in dict::entries(font_dict_data, None) {
+            match entry? {
+                dict::Entry::PrivateDictRange(r) => {
+                    range = Some(r);
+                }
+                // We store this matrix unnormalized since FreeType
+                // concatenates this with the top dict matrix (if present)
+                // before normalizing
+                dict::Entry::FontMatrix(matrix, upem) => font_matrix = Some((matrix, upem)),
+                _ => {}
+            }
+        }
+        Ok(Self {
+            private_dict_range: range.ok_or(Error::MissingPrivateDict)?,
+            font_matrix,
+        })
+    }
+}
+
 /// Entries that we parse from the Top DICT that are required to support
 /// charstring evaluation.
 #[derive(Clone, Default)]
@@ -347,6 +472,7 @@ struct TopDict<'a> {
     font_dicts: Index<'a>,
     fd_select: Option<FdSelect<'a>>,
     private_dict_range: Range<u32>,
+    font_matrix: Option<([Fixed; 6], i32)>,
     var_store: Option<ItemVariationStore<'a>>,
 }
 
@@ -371,6 +497,10 @@ impl<'a> TopDict<'a> {
                 dict::Entry::PrivateDictRange(range) => {
                     items.private_dict_range = range.start as u32..range.end as u32;
                 }
+                dict::Entry::FontMatrix(matrix, upem) => {
+                    // Store this matrix normalized since FT always applies normalization
+                    items.font_matrix = Some(normalize_font_matrix(matrix, upem));
+                }
                 dict::Entry::VariationStoreOffset(offset) if is_cff2 => {
                     // IVS is preceded by a 2 byte length, but ensure that
                     // we don't overflow
@@ -425,6 +555,181 @@ where
     }
 }
 
+fn transform(matrix: &[Fixed; 6], x: Fixed, y: Fixed) -> (Fixed, Fixed) {
+    (
+        matrix[0] * x + matrix[2] * y + matrix[4],
+        matrix[1] * x + matrix[3] * y + matrix[5],
+    )
+}
+
+/// Command sink adapter that applies a transform to hinted coordinates.
+struct HintedTransformingSink<'a, S> {
+    inner: &'a mut S,
+    matrix: [Fixed; 6],
+}
+
+impl<'a, S> HintedTransformingSink<'a, S> {
+    fn new(sink: &'a mut S, matrix: [Fixed; 6]) -> Self {
+        Self {
+            inner: sink,
+            matrix,
+        }
+    }
+
+    fn transform(&self, x: Fixed, y: Fixed) -> (Fixed, Fixed) {
+        // FreeType applies the transform to 26.6 values but we maintain
+        // values in 16.16 so convert, transform and then convert back
+        let (x, y) = transform(
+            &self.matrix,
+            Fixed::from_bits(x.to_bits() >> 10),
+            Fixed::from_bits(y.to_bits() >> 10),
+        );
+        (
+            Fixed::from_bits(x.to_bits() << 10),
+            Fixed::from_bits(y.to_bits() << 10),
+        )
+    }
+}
+
+impl<S: CommandSink> CommandSink for HintedTransformingSink<'_, S> {
+    fn hstem(&mut self, y: Fixed, dy: Fixed) {
+        self.inner.hstem(y, dy);
+    }
+
+    fn vstem(&mut self, x: Fixed, dx: Fixed) {
+        self.inner.vstem(x, dx);
+    }
+
+    fn hint_mask(&mut self, mask: &[u8]) {
+        self.inner.hint_mask(mask);
+    }
+
+    fn counter_mask(&mut self, mask: &[u8]) {
+        self.inner.counter_mask(mask);
+    }
+
+    fn move_to(&mut self, x: Fixed, y: Fixed) {
+        let (x, y) = self.transform(x, y);
+        self.inner.move_to(x, y);
+    }
+
+    fn line_to(&mut self, x: Fixed, y: Fixed) {
+        let (x, y) = self.transform(x, y);
+        self.inner.line_to(x, y);
+    }
+
+    fn curve_to(&mut self, cx1: Fixed, cy1: Fixed, cx2: Fixed, cy2: Fixed, x: Fixed, y: Fixed) {
+        let (cx1, cy1) = self.transform(cx1, cy1);
+        let (cx2, cy2) = self.transform(cx2, cy2);
+        let (x, y) = self.transform(x, y);
+        self.inner.curve_to(cx1, cy1, cx2, cy2, x, y);
+    }
+
+    fn close(&mut self) {
+        self.inner.close();
+    }
+}
+
+// Used for scaling sinks below
+const ONE_OVER_64: Fixed = Fixed::from_bits(0x400);
+
+/// Command sink adapter that applies both a transform and a scaling
+/// factor.
+struct ScalingTransformingSink<'a, S> {
+    inner: &'a mut S,
+    matrix: [Fixed; 6],
+    scale: Option<Fixed>,
+}
+
+impl<'a, S> ScalingTransformingSink<'a, S> {
+    fn new(sink: &'a mut S, matrix: [Fixed; 6], scale: Option<Fixed>) -> Self {
+        Self {
+            inner: sink,
+            matrix,
+            scale,
+        }
+    }
+
+    fn transform(&self, x: Fixed, y: Fixed) -> (Fixed, Fixed) {
+        // The following dance is necessary to exactly match FreeType's
+        // application of scaling factors. This seems to be the result
+        // of merging the contributed Adobe code while not breaking the
+        // FreeType public API.
+        //
+        // The first two steps apply to both scaled and unscaled outlines:
+        //
+        // 1. Multiply by 1/64
+        // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/psft.c#L284>
+        let ax = x * ONE_OVER_64;
+        let ay = y * ONE_OVER_64;
+        // 2. Truncate the bottom 10 bits. Combined with the division by 64,
+        // converts to font units.
+        // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/psobjs.c#L2219>
+        let bx = Fixed::from_bits(ax.to_bits() >> 10);
+        let by = Fixed::from_bits(ay.to_bits() >> 10);
+        // 3. Apply the transform. It must be done here to match FreeType.
+        let (cx, cy) = transform(&self.matrix, bx, by);
+        if let Some(scale) = self.scale {
+            // Scaled case:
+            // 4. Multiply by the original scale factor (to 26.6)
+            // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/cff/cffgload.c#L721>
+            let dx = cx * scale;
+            let dy = cy * scale;
+            // 5. Convert from 26.6 to 16.16
+            (
+                Fixed::from_bits(dx.to_bits() << 10),
+                Fixed::from_bits(dy.to_bits() << 10),
+            )
+        } else {
+            // Unscaled case:
+            // 4. Convert from integer to 16.16
+            (
+                Fixed::from_bits(cx.to_bits() << 16),
+                Fixed::from_bits(cy.to_bits() << 16),
+            )
+        }
+    }
+}
+
+impl<S: CommandSink> CommandSink for ScalingTransformingSink<'_, S> {
+    fn hstem(&mut self, y: Fixed, dy: Fixed) {
+        self.inner.hstem(y, dy);
+    }
+
+    fn vstem(&mut self, x: Fixed, dx: Fixed) {
+        self.inner.vstem(x, dx);
+    }
+
+    fn hint_mask(&mut self, mask: &[u8]) {
+        self.inner.hint_mask(mask);
+    }
+
+    fn counter_mask(&mut self, mask: &[u8]) {
+        self.inner.counter_mask(mask);
+    }
+
+    fn move_to(&mut self, x: Fixed, y: Fixed) {
+        let (x, y) = self.transform(x, y);
+        self.inner.move_to(x, y);
+    }
+
+    fn line_to(&mut self, x: Fixed, y: Fixed) {
+        let (x, y) = self.transform(x, y);
+        self.inner.line_to(x, y);
+    }
+
+    fn curve_to(&mut self, cx1: Fixed, cy1: Fixed, cx2: Fixed, cy2: Fixed, x: Fixed, y: Fixed) {
+        let (cx1, cy1) = self.transform(cx1, cy1);
+        let (cx2, cy2) = self.transform(cx2, cy2);
+        let (x, y) = self.transform(x, y);
+        self.inner.curve_to(cx1, cy1, cx2, cy2, x, y);
+    }
+
+    fn close(&mut self) {
+        self.inner.close();
+    }
+}
+
 /// Command sink adapter that applies a scaling factor.
 ///
 /// This assumes a 26.6 scaling factor packed into a Fixed and thus,
@@ -450,7 +755,7 @@ impl<'a, S> ScalingSink26Dot6<'a, S> {
         //
         // 1. Multiply by 1/64
         // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/psft.c#L284>
-        let a = coord * Fixed::from_bits(0x0400);
+        let a = coord * ONE_OVER_64;
         // 2. Truncate the bottom 10 bits. Combined with the division by 64,
         // converts to font units.
         // <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/psobjs.c#L2219>
@@ -511,6 +816,22 @@ impl<S: CommandSink> CommandSink for ScalingSink26Dot6<'_, S> {
     }
 }
 
+#[derive(Copy, Clone)]
+enum PendingElement {
+    Move([Fixed; 2]),
+    Line([Fixed; 2]),
+    Curve([Fixed; 6]),
+}
+
+impl PendingElement {
+    fn target_point(&self) -> [Fixed; 2] {
+        match self {
+            Self::Move(xy) | Self::Line(xy) => *xy,
+            Self::Curve([.., x, y]) => [*x, *y],
+        }
+    }
+}
+
 /// Command sink adapter that suppresses degenerate move and line commands.
 ///
 /// FreeType avoids emitting empty contours and zero length lines to prevent
@@ -520,9 +841,9 @@ impl<S: CommandSink> CommandSink for ScalingSink26Dot6<'_, S> {
 ///
 /// See <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/psaux/pshints.c#L1786>
 struct NopFilteringSink<'a, S> {
+    is_open: bool,
     start: Option<(Fixed, Fixed)>,
-    last: Option<(Fixed, Fixed)>,
-    pending_move: Option<(Fixed, Fixed)>,
+    pending_element: Option<PendingElement>,
     inner: &'a mut S,
 }
 
@@ -532,33 +853,37 @@ where
 {
     fn new(inner: &'a mut S) -> Self {
         Self {
+            is_open: false,
             start: None,
-            last: None,
-            pending_move: None,
+            pending_element: None,
             inner,
         }
     }
 
-    fn flush_pending_move(&mut self) {
-        if let Some((x, y)) = self.pending_move.take() {
-            if let Some((last_x, last_y)) = self.start {
-                if self.last != self.start {
-                    self.inner.line_to(last_x, last_y);
+    fn flush_pending(&mut self, for_close: bool) {
+        if let Some(pending) = self.pending_element.take() {
+            match pending {
+                PendingElement::Move([x, y]) => {
+                    if !for_close {
+                        self.is_open = true;
+                        self.inner.move_to(x, y);
+                        self.start = Some((x, y));
+                    }
+                }
+                PendingElement::Line([x, y]) => {
+                    if !for_close || self.start != Some((x, y)) {
+                        self.inner.line_to(x, y);
+                    }
+                }
+                PendingElement::Curve([cx0, cy0, cx1, cy1, x, y]) => {
+                    self.inner.curve_to(cx0, cy0, cx1, cy1, x, y);
                 }
             }
-            self.start = Some((x, y));
-            self.last = None;
-            self.inner.move_to(x, y);
         }
     }
 
     pub fn finish(&mut self) {
-        if let Some((x, y)) = self.start {
-            if self.last != self.start {
-                self.inner.line_to(x, y);
-            }
-            self.inner.close();
-        }
+        self.close();
     }
 }
 
@@ -583,36 +908,56 @@ where
     }
 
     fn move_to(&mut self, x: Fixed, y: Fixed) {
-        self.pending_move = Some((x, y));
+        self.pending_element = Some(PendingElement::Move([x, y]));
     }
 
     fn line_to(&mut self, x: Fixed, y: Fixed) {
-        if self.pending_move == Some((x, y)) {
+        // Omit the line if we're already at the given position
+        if self
+            .pending_element
+            .map(|element| element.target_point() == [x, y])
+            .unwrap_or_default()
+        {
             return;
         }
-        self.flush_pending_move();
-        if self.last == Some((x, y)) || (self.last.is_none() && self.start == Some((x, y))) {
-            return;
-        }
-        self.inner.line_to(x, y);
-        self.last = Some((x, y));
+        self.flush_pending(false);
+        self.pending_element = Some(PendingElement::Line([x, y]));
     }
 
     fn curve_to(&mut self, cx1: Fixed, cy1: Fixed, cx2: Fixed, cy2: Fixed, x: Fixed, y: Fixed) {
-        self.flush_pending_move();
-        self.last = Some((x, y));
-        self.inner.curve_to(cx1, cy1, cx2, cy2, x, y);
+        self.flush_pending(false);
+        self.pending_element = Some(PendingElement::Curve([cx1, cy1, cx2, cy2, x, y]));
     }
 
     fn close(&mut self) {
-        if self.pending_move.is_none() {
+        self.flush_pending(true);
+        if self.is_open {
             self.inner.close();
-            self.start = None;
-            self.last = None;
+            self.is_open = false;
         }
     }
 }
 
+/// Simple fixed point matrix multiplication with a scaling factor.
+///
+/// Note: this transforms the translation component of `b` by the upper 2x2 of
+/// `a`. This matches the offset transform FreeType uses when concatenating
+/// the matrices from the top and font dicts.
+///
+/// See <https://gitlab.freedesktop.org/freetype/freetype/-/blob/80a507a6b8e3d2906ad2c8ba69329bd2fb2a85ef/src/base/ftcalc.c#L719>
+fn matrix_mul_scaled(a: &[Fixed; 6], b: &[Fixed; 6], scaling: i32) -> [Fixed; 6] {
+    let val = Fixed::from_i32(scaling);
+    let xx = a[0].mul_div(b[0], val) + a[2].mul_div(b[1], val);
+    let yx = a[1].mul_div(b[0], val) + a[3].mul_div(b[1], val);
+    let xy = a[0].mul_div(b[2], val) + a[2].mul_div(b[3], val);
+    let yy = a[1].mul_div(b[2], val) + a[3].mul_div(b[3], val);
+    let x = b[4];
+    let y = b[5];
+    let dx = x.mul_div(a[0], val) + y.mul_div(a[2], val);
+    let dy = x.mul_div(a[1], val) + y.mul_div(a[3], val);
+    [xx, yx, xy, yy, dx, dy]
+}
+
 #[cfg(test)]
 mod tests {
     use super::{super::pen::SvgPen, *};
@@ -621,8 +966,10 @@ mod tests {
         prelude::{LocationRef, Size},
         MetadataProvider,
     };
+    use dict::Blues;
+    use font_test_data::bebuffer::BeBuffer;
     use raw::tables::cff2::Cff2;
-    use read_fonts::{test_helpers::BeBuffer, FontRef};
+    use read_fonts::FontRef;
 
     #[test]
     fn unscaled_scaling_sink_produces_integers() {
@@ -747,7 +1094,11 @@ mod tests {
         let font = FontRef::new(font_test_data::MATERIAL_ICONS_SUBSET).unwrap();
         let outlines = super::Outlines::new(&font).unwrap();
         assert!(outlines.top_dict.private_dict_range.is_empty());
-        assert!(outlines.private_dict_range(0).unwrap().is_empty());
+        assert!(outlines
+            .parse_font_dict(0)
+            .unwrap()
+            .private_dict_range
+            .is_empty());
     }
 
     /// Fuzzer caught add with overflow when computing subrs offset.
@@ -849,4 +1200,128 @@ mod tests {
             }
         }
     }
+
+    // We were overwriting family_other_blues with family_blues.
+    #[test]
+    fn capture_family_other_blues() {
+        let private_dict_data = &font_test_data::cff2::EXAMPLE[0x4f..=0xc0];
+        let store =
+            ItemVariationStore::read(FontData::new(&font_test_data::cff2::EXAMPLE[18..])).unwrap();
+        let coords = &[F2Dot14::from_f32(0.0)];
+        let blend_state = BlendState::new(store, coords, 0).unwrap();
+        let private_dict = PrivateDict::new(
+            FontData::new(private_dict_data),
+            0..private_dict_data.len(),
+            Some(blend_state),
+        )
+        .unwrap();
+        assert_eq!(
+            private_dict.hint_params.family_other_blues,
+            Blues::new([-249.0, -239.0].map(Fixed::from_f64).into_iter())
+        )
+    }
+
+    #[test]
+    fn implied_seac() {
+        let font = FontRef::new(font_test_data::CHARSTRING_PATH_OPS).unwrap();
+        let glyphs = font.outline_glyphs();
+        let gid = GlyphId::new(3);
+        assert_eq!(font.glyph_names().get(gid).unwrap(), "Scaron");
+        let glyph = glyphs.get(gid).unwrap();
+        let mut pen = SvgPen::new();
+        glyph
+            .draw((Size::unscaled(), LocationRef::default()), &mut pen)
+            .unwrap();
+        // This triggers the seac behavior in the endchar operator which
+        // loads an accent character followed by a base character. Ensure
+        // that we have a path to represent each by checking for two closepath
+        // commands.
+        assert_eq!(pen.to_string().chars().filter(|ch| *ch == 'Z').count(), 2);
+    }
+
+    const TRANSFORM: [Fixed; 6] = [
+        Fixed::ONE,
+        Fixed::ZERO,
+        // 0.167007446289062
+        Fixed::from_bits(10945),
+        Fixed::ONE,
+        Fixed::ZERO,
+        Fixed::ZERO,
+    ];
+
+    #[test]
+    fn hinted_transform_sink() {
+        // A few points taken from the test font in <https://github.com/googlefonts/fontations/issues/1581>
+        // Inputs and expected values extracted from FreeType
+        let input = [(383i32, 117i32), (450, 20), (555, -34), (683, -34)]
+            .map(|(x, y)| (Fixed::from_bits(x << 10), Fixed::from_bits(y << 10)));
+        let expected = [(403, 117i32), (453, 20), (549, -34), (677, -34)]
+            .map(|(x, y)| (Fixed::from_bits(x << 10), Fixed::from_bits(y << 10)));
+        let mut dummy = ();
+        let sink = HintedTransformingSink::new(&mut dummy, TRANSFORM);
+        let transformed = input.map(|(x, y)| sink.transform(x, y));
+        assert_eq!(transformed, expected);
+    }
+
+    #[test]
+    fn unhinted_scaled_transform_sink() {
+        // A few points taken from the test font in <https://github.com/googlefonts/fontations/issues/1581>
+        // Inputs and expected values extracted from FreeType
+        let input = [(150i32, 46i32), (176, 8), (217, -13), (267, -13)]
+            .map(|(x, y)| (Fixed::from_bits(x << 16), Fixed::from_bits(y << 16)));
+        let expected = [(404, 118i32), (453, 20), (550, -33), (678, -33)]
+            .map(|(x, y)| (Fixed::from_bits(x << 10), Fixed::from_bits(y << 10)));
+        let mut dummy = ();
+        let sink =
+            ScalingTransformingSink::new(&mut dummy, TRANSFORM, Some(Fixed::from_bits(167772)));
+        let transformed = input.map(|(x, y)| sink.transform(x, y));
+        assert_eq!(transformed, expected);
+    }
+
+    #[test]
+    fn unhinted_unscaled_transform_sink() {
+        // A few points taken from the test font in <https://github.com/googlefonts/fontations/issues/1581>
+        // Inputs and expected values extracted from FreeType
+        let input = [(150i32, 46i32), (176, 8), (217, -13), (267, -13)]
+            .map(|(x, y)| (Fixed::from_bits(x << 16), Fixed::from_bits(y << 16)));
+        let expected = [(158, 46i32), (177, 8), (215, -13), (265, -13)]
+            .map(|(x, y)| (Fixed::from_bits(x << 16), Fixed::from_bits(y << 16)));
+        let mut dummy = ();
+        let sink = ScalingTransformingSink::new(&mut dummy, TRANSFORM, None);
+        let transformed = input.map(|(x, y)| sink.transform(x, y));
+        assert_eq!(transformed, expected);
+    }
+
+    #[test]
+    fn fixed_matrix_mul() {
+        let a = [0.5, 0.75, -1.0, 2.0, 0.0, 0.0].map(Fixed::from_f64);
+        let b = [1.5, -1.0, 0.25, -1.0, 1.0, 2.0].map(Fixed::from_f64);
+        let expected = [1.75, -0.875, 1.125, -1.8125, -1.5, 4.75].map(Fixed::from_f64);
+        let result = matrix_mul_scaled(&a, &b, 1);
+        assert_eq!(result, expected);
+    }
+
+    /// See <https://github.com/googlefonts/fontations/issues/1638>
+    #[test]
+    fn nested_font_matrices() {
+        // Expected values extracted from FreeType debugging session
+        let font = FontRef::new(font_test_data::MATERIAL_ICONS_SUBSET_MATRIX).unwrap();
+        let outlines = Outlines::from_cff(&font, 512).unwrap();
+        // Check the normalized top dict matrix
+        let (top_matrix, top_upem) = outlines.top_dict.font_matrix.unwrap();
+        let expected_top_matrix = [65536, 0, 5604, 65536, 0, 0].map(Fixed::from_bits);
+        assert_eq!(top_matrix, expected_top_matrix);
+        assert_eq!(top_upem, 512);
+        // Check the unnormalized font dict matrix
+        let (sub_matrix, sub_upem) = outlines.parse_font_dict(0).unwrap().font_matrix.unwrap();
+        let expected_sub_matrix = [327680, 0, 0, 327680, 0, 0].map(Fixed::from_bits);
+        assert_eq!(sub_matrix, expected_sub_matrix);
+        assert_eq!(sub_upem, 10);
+        // Check the normalized combined matrix
+        let subfont = outlines.subfont(0, Some(24.0), &[]).unwrap();
+        let expected_combined_matrix = [65536, 0, 5604, 65536, 0, 0].map(Fixed::from_bits);
+        assert_eq!(subfont.font_matrix.unwrap(), expected_combined_matrix);
+        // Check the final scale
+        assert_eq!(subfont.scale.unwrap().to_bits(), 98304);
+    }
 }
