@@ -27,8 +27,6 @@ impl AllowCredentials {
 
     /// Allow credentials for some requests, based on a given predicate
     ///
-    /// The first argument to the predicate is the request origin.
-    ///
     /// See [`CorsLayer::allow_credentials`] for more details.
     ///
     /// [`CorsLayer::allow_credentials`]: super::CorsLayer::allow_credentials
@@ -57,7 +55,7 @@ impl AllowCredentials {
             AllowCredentialsInner::Predicate(c) => c(origin?, parts),
         };
 
-        allow_creds.then_some((header::ACCESS_CONTROL_ALLOW_CREDENTIALS, TRUE))
+        allow_creds.then(|| (header::ACCESS_CONTROL_ALLOW_CREDENTIALS, TRUE))
     }
 }
 
