@@ -4,11 +4,10 @@ use indexmap::{IndexMap, IndexSet};
 use std::{
     borrow::Cow,
     collections::{HashMap, HashSet},
-    convert::Infallible,
     mem,
     ops::Deref,
 };
-use wasm_encoder::{reencode::Reencode, Encode, EntityType, RawCustomSection};
+use wasm_encoder::{Encode, EntityType, Instruction, RawCustomSection};
 use wasmparser::*;
 
 const PAGE_SIZE: i32 = 64 * 1024;
@@ -18,9 +17,9 @@ const PAGE_SIZE: i32 = 64 * 1024;
 ///
 /// This internally performs a "gc" pass after removing exports to ensure that
 /// the resulting module imports the minimal set of functions necessary.
-pub fn run(
+pub fn run<T>(
     wasm: &[u8],
-    required: &IndexSet<String>,
+    required: &IndexMap<String, T>,
     main_module_realloc: Option<&str>,
 ) -> Result<Vec<u8>> {
     assert!(!required.is_empty());
@@ -30,71 +29,89 @@ pub fn run(
 
     // Make sure that all required names are present in the module, and then
     // remove all names that are not required.
-    for name in required {
+    for (name, _ty) in required {
         if !module.exports.contains_key(name.as_str()) {
             bail!("adapter module does not have export `{name}`")
         }
     }
     let mut not_required = IndexSet::new();
     for name in module.exports.keys().copied() {
-        if !required.contains(name) {
+        // If we need `name` then we also need cabi_post_`name`:
+        let name = if let Some(suffix) = name.strip_prefix("cabi_post_") {
+            suffix
+        } else {
+            name
+        };
+
+        if !required.contains_key(name) && !always_keep(name) {
             not_required.insert(name);
         }
     }
     for name in not_required {
-        module.exports.swap_remove(name);
+        module.exports.remove(name);
     }
     assert!(!module.exports.is_empty());
     module.liveness()?;
     module.encode(main_module_realloc)
 }
 
+fn always_keep(name: &str) -> bool {
+    match name {
+        // Explicitly keep `cabi_realloc` if it's there in case an interface
+        // needs it for a lowering.
+        "cabi_realloc" | "cabi_import_realloc" | "cabi_export_realloc" => true,
+        _ => false,
+    }
+}
+
 /// This function generates a Wasm function body which implements `cabi_realloc` in terms of `memory.grow`.  It
 /// only accepts new, page-sized allocations.
 fn realloc_via_memory_grow() -> wasm_encoder::Function {
+    use wasm_encoder::Instruction::*;
+
     let mut func = wasm_encoder::Function::new([(1, wasm_encoder::ValType::I32)]);
 
     // Assert `old_ptr` is null.
-    func.instructions().i32_const(0);
-    func.instructions().local_get(0);
-    func.instructions().i32_ne();
-    func.instructions().if_(wasm_encoder::BlockType::Empty);
-    func.instructions().unreachable();
-    func.instructions().end();
+    func.instruction(&I32Const(0));
+    func.instruction(&LocalGet(0));
+    func.instruction(&I32Ne);
+    func.instruction(&If(wasm_encoder::BlockType::Empty));
+    func.instruction(&Unreachable);
+    func.instruction(&End);
 
     // Assert `old_len` is zero.
-    func.instructions().i32_const(0);
-    func.instructions().local_get(1);
-    func.instructions().i32_ne();
-    func.instructions().if_(wasm_encoder::BlockType::Empty);
-    func.instructions().unreachable();
-    func.instructions().end();
+    func.instruction(&I32Const(0));
+    func.instruction(&LocalGet(1));
+    func.instruction(&I32Ne);
+    func.instruction(&If(wasm_encoder::BlockType::Empty));
+    func.instruction(&Unreachable);
+    func.instruction(&End);
 
     // Assert `new_len` is equal to the page size (which is the only value we currently support)
     // Note: we could easily support arbitrary multiples of PAGE_SIZE here if the need arises.
-    func.instructions().i32_const(PAGE_SIZE);
-    func.instructions().local_get(3);
-    func.instructions().i32_ne();
-    func.instructions().if_(wasm_encoder::BlockType::Empty);
-    func.instructions().unreachable();
-    func.instructions().end();
+    func.instruction(&I32Const(PAGE_SIZE));
+    func.instruction(&LocalGet(3));
+    func.instruction(&I32Ne);
+    func.instruction(&If(wasm_encoder::BlockType::Empty));
+    func.instruction(&Unreachable);
+    func.instruction(&End);
 
     // Grow the memory by 1 page.
-    func.instructions().i32_const(1);
-    func.instructions().memory_grow(0);
-    func.instructions().local_tee(4);
+    func.instruction(&I32Const(1));
+    func.instruction(&MemoryGrow(0));
+    func.instruction(&LocalTee(4));
 
     // Test if the return value of the growth was -1 and, if so, trap due to a failed allocation.
-    func.instructions().i32_const(-1);
-    func.instructions().i32_eq();
-    func.instructions().if_(wasm_encoder::BlockType::Empty);
-    func.instructions().unreachable();
-    func.instructions().end();
+    func.instruction(&I32Const(-1));
+    func.instruction(&I32Eq);
+    func.instruction(&If(wasm_encoder::BlockType::Empty));
+    func.instruction(&Unreachable);
+    func.instruction(&End);
 
-    func.instructions().local_get(4);
-    func.instructions().i32_const(16);
-    func.instructions().i32_shl();
-    func.instructions().end();
+    func.instruction(&LocalGet(4));
+    func.instruction(&I32Const(16));
+    func.instruction(&I32Shl);
+    func.instruction(&End);
 
     func
 }
@@ -112,39 +129,38 @@ fn allocate_stack_via_realloc(
     sp: u32,
     allocation_state: Option<u32>,
 ) -> wasm_encoder::Function {
+    use wasm_encoder::Instruction::*;
+
     let mut func = wasm_encoder::Function::new([]);
 
     if let Some(allocation_state) = allocation_state {
         // This means we're lazily allocating the stack, keeping track of state via `$allocation_state`
-        func.instructions().global_get(allocation_state);
-        func.instructions()
-            .i32_const(StackAllocationState::Unallocated as _);
-        func.instructions().i32_eq();
-        func.instructions().if_(wasm_encoder::BlockType::Empty);
-        func.instructions()
-            .i32_const(StackAllocationState::Allocating as _);
-        func.instructions().global_set(allocation_state);
+        func.instruction(&GlobalGet(allocation_state));
+        func.instruction(&I32Const(StackAllocationState::Unallocated as _));
+        func.instruction(&I32Eq);
+        func.instruction(&If(wasm_encoder::BlockType::Empty));
+        func.instruction(&I32Const(StackAllocationState::Allocating as _));
+        func.instruction(&GlobalSet(allocation_state));
         // We could also set `sp` to zero here to ensure the yet-to-be-allocated stack is empty.  However, we
         // assume it defaults to zero anyway, in which case setting it would be redundant.
     }
 
-    func.instructions().i32_const(0);
-    func.instructions().i32_const(0);
-    func.instructions().i32_const(8);
-    func.instructions().i32_const(PAGE_SIZE);
-    func.instructions().call(realloc_index);
-    func.instructions().i32_const(PAGE_SIZE);
-    func.instructions().i32_add();
-    func.instructions().global_set(sp);
+    func.instruction(&I32Const(0));
+    func.instruction(&I32Const(0));
+    func.instruction(&I32Const(8));
+    func.instruction(&I32Const(PAGE_SIZE));
+    func.instruction(&Call(realloc_index));
+    func.instruction(&I32Const(PAGE_SIZE));
+    func.instruction(&I32Add);
+    func.instruction(&GlobalSet(sp));
 
     if let Some(allocation_state) = allocation_state {
-        func.instructions()
-            .i32_const(StackAllocationState::Allocated as _);
-        func.instructions().global_set(allocation_state);
-        func.instructions().end();
+        func.instruction(&I32Const(StackAllocationState::Allocated as _));
+        func.instruction(&GlobalSet(allocation_state));
+        func.instruction(&End);
     }
 
-    func.instructions().end();
+    func.instruction(&End);
 
     func
 }
@@ -299,7 +315,7 @@ impl<'a> Module<'a> {
                         self.funcs.push(Func {
                             // Specify a dummy definition to get filled in later
                             // when parsing the code section.
-                            def: Definition::Local(FunctionBody::new(BinaryReader::new(&[], 0))),
+                            def: Definition::Local(FunctionBody::new(0, &[])),
                             ty,
                         });
                     }
@@ -313,25 +329,48 @@ impl<'a> Module<'a> {
 
                 // Ignore all custom sections except for the `name` and
                 // `producers` sections which we parse, but ignore errors within.
-                Payload::CustomSection(s) => match s.as_known() {
-                    KnownCustom::Name(s) => drop(self.parse_name_section(s)),
-                    KnownCustom::Producers(_) => drop(self.parse_producers_section(&s)),
-                    _ => {}
-                },
+                Payload::CustomSection(s) => {
+                    if s.name() == "name" {
+                        drop(self.parse_name_section(&s));
+                    }
+                    if s.name() == "producers" {
+                        drop(self.parse_producers_section(&s));
+                    }
+                }
 
-                // sections that shouldn't appear in the specially-crafted core
-                // wasm adapter self we're processing
-                other => match other.as_section() {
-                    Some((id, _)) => bail!("unsupported section `{}` in adapter", id),
-                    None => bail!("unsupported payload in adapter"),
-                },
+                // sections that shouldn't appear in the specially-crafted core wasm
+                // adapter self we're processing
+                Payload::DataCountSection { .. }
+                | Payload::ElementSection(_)
+                | Payload::DataSection(_)
+                | Payload::StartSection { .. }
+                | Payload::TagSection(_)
+                | Payload::UnknownSection { .. } => {
+                    bail!("unsupported section found in adapter module")
+                }
+
+                // component-model related things that shouldn't show up
+                Payload::ModuleSection { .. }
+                | Payload::ComponentSection { .. }
+                | Payload::InstanceSection(_)
+                | Payload::ComponentInstanceSection(_)
+                | Payload::ComponentAliasSection(_)
+                | Payload::ComponentCanonicalSection(_)
+                | Payload::ComponentStartSection { .. }
+                | Payload::ComponentImportSection(_)
+                | Payload::CoreTypeSection(_)
+                | Payload::ComponentExportSection(_)
+                | Payload::ComponentTypeSection(_) => {
+                    bail!("component section found in adapter module")
+                }
             }
         }
 
         Ok(())
     }
 
-    fn parse_name_section(&mut self, section: NameSectionReader<'a>) -> Result<()> {
+    fn parse_name_section(&mut self, section: &CustomSectionReader<'a>) -> Result<()> {
+        let section = NameSectionReader::new(section.data(), section.data_offset());
         for s in section {
             match s? {
                 Name::Function(map) => {
@@ -447,7 +486,17 @@ impl<'a> Module<'a> {
 
     fn heapty(&mut self, ty: HeapType) {
         match ty {
-            HeapType::Abstract { .. } => {}
+            HeapType::Func
+            | HeapType::Extern
+            | HeapType::Any
+            | HeapType::None
+            | HeapType::NoExtern
+            | HeapType::NoFunc
+            | HeapType::Eq
+            | HeapType::Struct
+            | HeapType::Array
+            | HeapType::I31
+            | HeapType::Exn => {}
             HeapType::Concrete(i) => self.ty(i.as_module_index().unwrap()),
         }
     }
@@ -513,8 +562,10 @@ impl<'a> Module<'a> {
         for (i, ty) in self.live_types() {
             map.types.push(i);
 
-            let ty = map.func_type(ty.clone())?;
-            types.ty().func_type(&ty);
+            types.function(
+                ty.params().iter().map(|t| map.valty(*t)),
+                ty.results().iter().map(|t| map.valty(*t)),
+            );
 
             // Keep track of the "empty type" to see if we can reuse an
             // existing one or one needs to be injected if a `start`
@@ -527,7 +578,12 @@ impl<'a> Module<'a> {
         let mut num_memories = 0;
         for (i, mem) in self.live_memories() {
             map.memories.push(i);
-            let ty = map.memory_type(mem.ty);
+            let ty = wasm_encoder::MemoryType {
+                minimum: mem.ty.initial,
+                maximum: mem.ty.maximum,
+                shared: mem.ty.shared,
+                memory64: mem.ty.memory64,
+            };
             match &mem.def {
                 Definition::Import(m, n) => {
                     imports.import(m, n, ty);
@@ -541,7 +597,11 @@ impl<'a> Module<'a> {
 
         for (i, table) in self.live_tables() {
             map.tables.push(i);
-            let ty = map.table_type(table.ty)?;
+            let ty = wasm_encoder::TableType {
+                minimum: table.ty.initial,
+                maximum: table.ty.maximum,
+                element_type: map.refty(table.ty.element_type),
+            };
             match &table.def {
                 Definition::Import(m, n) => {
                     imports.import(m, n, ty);
@@ -554,14 +614,18 @@ impl<'a> Module<'a> {
 
         for (i, global) in self.live_globals() {
             map.globals.push(i);
-            let ty = map.global_type(global.ty)?;
+            let ty = wasm_encoder::GlobalType {
+                mutable: global.ty.mutable,
+                val_type: map.valty(global.ty.content_type),
+            };
             match &global.def {
                 Definition::Import(m, n) => {
                     imports.import(m, n, ty);
                 }
                 Definition::Local(init) => {
-                    let init = &map.const_expr(init.clone())?;
-                    globals.global(ty, &init);
+                    let mut bytes = map.operators(init.get_binary_reader())?;
+                    assert_eq!(bytes.pop(), Some(0xb));
+                    globals.global(ty, &wasm_encoder::ConstExpr::raw(bytes));
                 }
             }
         }
@@ -608,7 +672,7 @@ impl<'a> Module<'a> {
 
         let add_realloc_type = |types: &mut wasm_encoder::TypeSection| {
             let type_index = types.len();
-            types.ty().function(
+            types.function(
                 [
                     wasm_encoder::ValType::I32,
                     wasm_encoder::ValType::I32,
@@ -622,7 +686,7 @@ impl<'a> Module<'a> {
 
         let add_empty_type = |types: &mut wasm_encoder::TypeSection| {
             let type_index = types.len();
-            types.ty().function([], []);
+            types.function([], []);
             type_index
         };
 
@@ -692,27 +756,26 @@ impl<'a> Module<'a> {
             .collect::<HashSet<_>>();
 
         for (i, func) in self.live_funcs() {
-            let body = match &func.def {
+            let mut body = match &func.def {
                 Definition::Import(..) => continue,
-                Definition::Local(body) => body,
+                Definition::Local(body) => body.get_binary_reader(),
             };
-
-            match (lazy_stack_init_index, exported_funcs.contains(&i)) {
-                // Prepend an `allocate_stack` call to all exports if we're
-                // lazily allocating the stack.
-                (Some(lazy_stack_init_index), true) => {
-                    let mut func = map.new_function_with_parsed_locals(&body)?;
-                    func.instructions().call(lazy_stack_init_index);
-                    let mut reader = body.get_operators_reader()?;
-                    while !reader.eof() {
-                        func.instruction(&map.parse_instruction(&mut reader)?);
-                    }
-                    code.function(&func);
-                }
-                _ => {
-                    map.parse_function_body(&mut code, body.clone())?;
-                }
+            let mut locals = Vec::new();
+            for _ in 0..body.read_var_u32()? {
+                let cnt = body.read_var_u32()?;
+                let ty = body.read()?;
+                locals.push((cnt, map.valty(ty)));
             }
+            // Prepend an `allocate_stack` call to all exports if we're lazily allocating the stack.
+            if let (Some(lazy_stack_init_index), true) =
+                (lazy_stack_init_index, exported_funcs.contains(&i))
+            {
+                Instruction::Call(lazy_stack_init_index).encode(&mut map.buf);
+            }
+            let bytes = map.operators(body)?;
+            let mut func = wasm_encoder::Function::new(locals);
+            func.raw(bytes);
+            code.function(&func);
         }
 
         if lazy_stack_init_index.is_some() {
@@ -751,7 +814,7 @@ impl<'a> Module<'a> {
                 // Generate a function type for this start function, adding a new
                 // function type to the module if necessary.
                 let empty_type = empty_type.unwrap_or_else(|| {
-                    types.ty().function([], []);
+                    types.function([], []);
                     types.len() - 1
                 });
                 funcs.function(empty_type);
@@ -929,7 +992,7 @@ impl<'a> Module<'a> {
 // While not exactly the most robust solution this should work well enough for
 // now.
 macro_rules! define_visit {
-    ($(@$p:ident $op:ident $({ $($arg:ident: $argty:ty),* })? => $visit:ident ($($ann:tt)*))*) => {
+    ($(@$p:ident $op:ident $({ $($arg:ident: $argty:ty),* })? => $visit:ident)*) => {
         $(
             fn $visit(&mut self $(, $($arg: $argty),*)?)  {
                 $(
@@ -966,6 +1029,8 @@ macro_rules! define_visit {
     (mark_live $self:ident $arg:ident lanes) => {};
     (mark_live $self:ident $arg:ident flags) => {};
     (mark_live $self:ident $arg:ident value) => {};
+    (mark_live $self:ident $arg:ident mem_byte) => {};
+    (mark_live $self:ident $arg:ident table_byte) => {};
     (mark_live $self:ident $arg:ident local_index) => {};
     (mark_live $self:ident $arg:ident relative_depth) => {};
     (mark_live $self:ident $arg:ident tag_index) => {};
@@ -978,26 +1043,13 @@ macro_rules! define_visit {
     (mark_live $self:ident $arg:ident field_index) => {};
     (mark_live $self:ident $arg:ident from_type_nullable) => {};
     (mark_live $self:ident $arg:ident to_type_nullable) => {};
-    (mark_live $self:ident $arg:ident ordering) => {};
     (mark_live $self:ident $arg:ident try_table) => {unimplemented!();};
-    (mark_live $self:ident $arg:ident argument_index) => {};
-    (mark_live $self:ident $arg:ident result_index) => {};
-    (mark_live $self:ident $arg:ident cont_type_index) => {};
-    (mark_live $self:ident $arg:ident resume_table) => {unimplemented!();};
 }
 
 impl<'a> VisitOperator<'a> for Module<'a> {
     type Output = ();
 
-    fn simd_visitor(&mut self) -> Option<&mut dyn VisitSimdOperator<'a, Output = Self::Output>> {
-        Some(self)
-    }
-
-    wasmparser::for_each_visit_operator!(define_visit);
-}
-
-impl<'a> VisitSimdOperator<'a> for Module<'a> {
-    wasmparser::for_each_visit_simd_operator!(define_visit);
+    wasmparser::for_each_operator!(define_visit);
 }
 
 /// Helper function to filter `iter` based on the `live` set, yielding an
@@ -1023,28 +1075,192 @@ struct Encoder {
     memories: Remap,
     globals: Remap,
     tables: Remap,
+    buf: Vec<u8>,
 }
 
-impl Reencode for Encoder {
-    type Error = Infallible;
+impl Encoder {
+    fn operators(&mut self, mut reader: BinaryReader<'_>) -> Result<Vec<u8>> {
+        while !reader.eof() {
+            reader.visit_operator(self)?;
+        }
+        Ok(mem::take(&mut self.buf))
+    }
 
-    fn type_index(&mut self, i: u32) -> u32 {
-        self.types.remap(i)
+    fn memarg(&self, ty: MemArg) -> wasm_encoder::MemArg {
+        wasm_encoder::MemArg {
+            offset: ty.offset,
+            align: ty.align.into(),
+            memory_index: self.memories.remap(ty.memory),
+        }
     }
-    fn function_index(&mut self, i: u32) -> u32 {
-        self.funcs.remap(i)
+
+    fn blockty(&self, ty: BlockType) -> wasm_encoder::BlockType {
+        match ty {
+            BlockType::Empty => wasm_encoder::BlockType::Empty,
+            BlockType::Type(ty) => wasm_encoder::BlockType::Result(self.valty(ty)),
+            BlockType::FuncType(ty) => wasm_encoder::BlockType::FunctionType(self.types.remap(ty)),
+        }
     }
-    fn memory_index(&mut self, i: u32) -> u32 {
-        self.memories.remap(i)
+
+    fn valty(&self, ty: wasmparser::ValType) -> wasm_encoder::ValType {
+        match ty {
+            wasmparser::ValType::I32 => wasm_encoder::ValType::I32,
+            wasmparser::ValType::I64 => wasm_encoder::ValType::I64,
+            wasmparser::ValType::F32 => wasm_encoder::ValType::F32,
+            wasmparser::ValType::F64 => wasm_encoder::ValType::F64,
+            wasmparser::ValType::V128 => wasm_encoder::ValType::V128,
+            wasmparser::ValType::Ref(rt) => wasm_encoder::ValType::Ref(self.refty(rt)),
+        }
     }
-    fn global_index(&mut self, i: u32) -> u32 {
-        self.globals.remap(i)
+
+    fn refty(&self, rt: wasmparser::RefType) -> wasm_encoder::RefType {
+        wasm_encoder::RefType {
+            nullable: rt.is_nullable(),
+            heap_type: self.heapty(rt.heap_type()),
+        }
     }
-    fn table_index(&mut self, i: u32) -> u32 {
-        self.tables.remap(i)
+
+    fn heapty(&self, ht: wasmparser::HeapType) -> wasm_encoder::HeapType {
+        match ht {
+            HeapType::Func => wasm_encoder::HeapType::Func,
+            HeapType::Extern => wasm_encoder::HeapType::Extern,
+            HeapType::Any => wasm_encoder::HeapType::Any,
+            HeapType::None => wasm_encoder::HeapType::None,
+            HeapType::NoExtern => wasm_encoder::HeapType::NoExtern,
+            HeapType::NoFunc => wasm_encoder::HeapType::NoFunc,
+            HeapType::Eq => wasm_encoder::HeapType::Eq,
+            HeapType::Struct => wasm_encoder::HeapType::Struct,
+            HeapType::Array => wasm_encoder::HeapType::Array,
+            HeapType::I31 => wasm_encoder::HeapType::I31,
+            HeapType::Exn => wasm_encoder::HeapType::Exn,
+            HeapType::Concrete(idx) => {
+                wasm_encoder::HeapType::Concrete(self.types.remap(idx.as_module_index().unwrap()))
+            }
+        }
     }
 }
 
+// This is a helper macro to translate all `wasmparser` instructions to
+// `wasm-encoder` instructions without having to list out every single
+// instruction itself.
+//
+// The general goal of this macro is to have O(unique instruction payload)
+// number of cases while also simultaneously adapting between the styles of
+// wasmparser and wasm-encoder.
+macro_rules! define_encode {
+    ($(@$p:ident $op:ident $({ $($arg:ident: $argty:ty),* })? => $visit:ident)*) => {
+        $(
+            fn $visit(&mut self $(, $($arg: $argty),*)?)  {
+                #[allow(unused_imports)]
+                use wasm_encoder::Instruction::*;
+                $(
+                    $(
+                        let $arg = define_encode!(map self $arg $arg);
+                    )*
+                )?
+                let insn = define_encode!(mk $op $($($arg)*)?);
+                insn.encode(&mut self.buf);
+            }
+        )*
+    };
+
+    // No-payload instructions are named the same in wasmparser as they are in
+    // wasm-encoder
+    (mk $op:ident) => ($op);
+
+    // Instructions which need "special care" to map from wasmparser to
+    // wasm-encoder
+    (mk BrTable $arg:ident) => ({
+        BrTable($arg.0, $arg.1)
+    });
+    (mk CallIndirect $ty:ident $table:ident $table_byte:ident) => ({
+        let _ = $table_byte;
+        CallIndirect { ty: $ty, table: $table }
+    });
+    (mk ReturnCallIndirect $ty:ident $table:ident) => (
+        ReturnCallIndirect { ty: $ty, table: $table }
+    );
+    (mk MemorySize $mem:ident $mem_byte:ident) => ({
+        let _ = $mem_byte;
+        MemorySize($mem)
+    });
+    (mk MemoryGrow $mem:ident $mem_byte:ident) => ({
+        let _ = $mem_byte;
+        MemoryGrow($mem)
+    });
+    (mk TryTable $try_table:ident) => ({
+        let _ = $try_table;
+        unimplemented_try_table()
+    });
+    (mk I32Const $v:ident) => (I32Const($v));
+    (mk I64Const $v:ident) => (I64Const($v));
+    (mk F32Const $v:ident) => (F32Const(f32::from_bits($v.bits())));
+    (mk F64Const $v:ident) => (F64Const(f64::from_bits($v.bits())));
+    (mk V128Const $v:ident) => (V128Const($v.i128()));
+
+    // Catch-all for the translation of one payload argument which is typically
+    // represented as a tuple-enum in wasm-encoder.
+    (mk $op:ident $arg:ident) => ($op($arg));
+
+    // Catch-all of everything else where the wasmparser fields are simply
+    // translated to wasm-encoder fields.
+    (mk $op:ident $($arg:ident)*) => ($op { $($arg),* });
+
+    // Individual cases of mapping one argument type to another, similar to the
+    // `define_visit` macro above.
+    (map $self:ident $arg:ident memarg) => {$self.memarg($arg)};
+    (map $self:ident $arg:ident blockty) => {$self.blockty($arg)};
+    (map $self:ident $arg:ident hty) => {$self.heapty($arg)};
+    (map $self:ident $arg:ident from_ref_type) => {$self.refty($arg)};
+    (map $self:ident $arg:ident to_ref_type) => {$self.refty($arg)};
+    (map $self:ident $arg:ident tag_index) => {$arg};
+    (map $self:ident $arg:ident relative_depth) => {$arg};
+    (map $self:ident $arg:ident function_index) => {$self.funcs.remap($arg)};
+    (map $self:ident $arg:ident global_index) => {$self.globals.remap($arg)};
+    (map $self:ident $arg:ident mem) => {$self.memories.remap($arg)};
+    (map $self:ident $arg:ident src_mem) => {$self.memories.remap($arg)};
+    (map $self:ident $arg:ident dst_mem) => {$self.memories.remap($arg)};
+    (map $self:ident $arg:ident table) => {$self.tables.remap($arg)};
+    (map $self:ident $arg:ident table_index) => {$self.tables.remap($arg)};
+    (map $self:ident $arg:ident src_table) => {$self.tables.remap($arg)};
+    (map $self:ident $arg:ident dst_table) => {$self.tables.remap($arg)};
+    (map $self:ident $arg:ident type_index) => {$self.types.remap($arg)};
+    (map $self:ident $arg:ident array_type_index) => {$self.types.remap($arg)};
+    (map $self:ident $arg:ident array_type_index_dst) => {$self.types.remap($arg)};
+    (map $self:ident $arg:ident array_type_index_src) => {$self.types.remap($arg)};
+    (map $self:ident $arg:ident struct_type_index) => {$self.types.remap($arg)};
+    (map $self:ident $arg:ident ty) => {$self.valty($arg)};
+    (map $self:ident $arg:ident local_index) => {$arg};
+    (map $self:ident $arg:ident lane) => {$arg};
+    (map $self:ident $arg:ident lanes) => {$arg};
+    (map $self:ident $arg:ident elem_index) => {$arg};
+    (map $self:ident $arg:ident data_index) => {$arg};
+    (map $self:ident $arg:ident array_elem_index) => {$arg};
+    (map $self:ident $arg:ident array_data_index) => {$arg};
+    (map $self:ident $arg:ident table_byte) => {$arg};
+    (map $self:ident $arg:ident mem_byte) => {$arg};
+    (map $self:ident $arg:ident value) => {$arg};
+    (map $self:ident $arg:ident array_size) => {$arg};
+    (map $self:ident $arg:ident field_index) => {$arg};
+    (map $self:ident $arg:ident from_type_nullable) => {$arg};
+    (map $self:ident $arg:ident to_type_nullable) => {$arg};
+    (map $self:ident $arg:ident try_table) => {$arg};
+    (map $self:ident $arg:ident targets) => ((
+        $arg.targets().map(|i| i.unwrap()).collect::<Vec<_>>().into(),
+        $arg.default(),
+    ));
+}
+
+fn unimplemented_try_table() -> wasm_encoder::Instruction<'static> {
+    unimplemented!()
+}
+
+impl<'a> VisitOperator<'a> for Encoder {
+    type Output = ();
+
+    wasmparser::for_each_operator!(define_encode);
+}
+
 // Minimal definition of a bit vector necessary for the liveness calculations
 // above.
 mod bitvec {
